
/*===============================<o>=====================================

  This software was written by Ian L. Kaplan, Chief Fat Bear, Bear Products
  International.  Use of this software, for any purpose, is granted on two
  conditions: 

   1. This copyright notice must be included with the software
      or any software derived from it.

   2. The risk of using this software is accepted by the user.  No
      warranty as to its usefulness or functionality is provided
      or implied.  The author and Bear Products International provides
      no support.

   May 1996

 ===============================<o>=====================================*/

//
// sier.cpp
//
// Class functions for pyramid, Sierpinski and debug_sierpinski classes.
//


#include <stdlib.h>
#include <stdio.h>
#include "sier.h"


//
// pyramid constructor
//
// Constructor for the pyramid class.  This constructor is passed the
// three_D_point for the pyramid apex.  It calculates the four corners
// of the pyramid and initializes the pyramid data structure.
//
// For example, consider the case where the pyramid has a height of
// the pyramid is two units and each side of the base is two units.
// If the bottom left corner of the pyramid is placed at 0,0,0 with
// the base on the x-z plane, the five points that make up the pyramid
// will be:
//
//    apex 1,2,1
//              _
//    A    0,0,0 |
//    B    2,0,0 |- base
//    C    2,0,2 |
//    D    0,0,2_|
//
// If the apex is at (Ax, Ay, Az), the other points will be at:
//
//    apex = (Ax, Ay, Az);
//    A = (Ax-1, Ay-2, Az-1)
//    B = (Ax+1, Ay-2, Az-1)
//    C = (Ax+1, Ay-2, Az+1)
//    D = (Ax-1, Ay-2, Az+1)
//
// The faces of the pyramid will be:
//
// side1 = A = (Ax-1, Ay-2, Az-1)
//         B = (Ax+1, Ay-2, Az-1)
//         apex = (Ax, Ay, Az)
//
// side2 = B = (Ax+1, Ay-2, Az-1)
//         C = (Ax+1, Ay-2, Az+1)
//         apex = (Ax, Ay, Az)
//
// side3 = C = (Ax+1, Ay-2, Az+1)
//         D = (Ax-1, Ay-2, Az+1)
//         apex = (Ax, Ay, Az)
//
// side4 = D = (Ax-1, Ay-2, Az+1)
//         A = (Ax-1, Ay-2, Az-1)
//         apex = (Ax, Ay, Az)
//
// bottom = A = (Ax-1, Ay-2, Az-1)
//          B = (Ax+1, Ay-2, Az-1)
//          C = (Ax+1, Ay-2, Az+1)
//          D = (Ax-1, Ay-2, Az+1)
//
// In this case the height is 2.  Any factor of two could be used
// although the calculation for the height of the Sierpinski pyramid
// would have to be adjusted.  However, since graphics objects can easily
// be scaled using openGL or a rendering package, there is no point
// in changing the height value.
//
pyramid::pyramid( three_D_point apex )
{
  int Ax, Ay, Az;

  Ax = apex.getX();
  Ay = apex.getY();
  Az = apex.getZ();
  pyramid_points[ Apex ] = apex;
  pyramid_points[ A ].set_point(Ax-1, Ay-2, Az-1);
  pyramid_points[ B ].set_point(Ax+1, Ay-2, Az-1);
  pyramid_points[ C ].set_point(Ax+1, Ay-2, Az+1);
  pyramid_points[ D ].set_point(Ax-1, Ay-2, Az+1);
} // pyramid



/*
 * pyramid::pyr_get_faces.
 *

   This function is passed a pointer to an integer array.  It stores
   the coordinates for the faces of a pyramid in the array.

   The pyramid is composed of five faces: four triangular faces and the
   square base.  Each triangular face shares the apex as a point.
   Following the scheme outlined in the comment for the pyramid constructor,
   above, these faces are:

     tri1 = apex, A, B
     tri2 = apex, B, C
     tri3 = apex, C, D
     tri4 = apex, D, A
     tribase = A, B, C, D

   To conform to the DXF file format's 3DFACE entity and to make the
   face data structure consistent, each face has four points.  Since
   all but the last (square) face contain only three points, the last 
   point is repeated.

 */
void pyramid::pyr_get_faces( int *&list )
{
  typedef enum { tri1 = 0, tri2, tri3, tri4, tribase } face_names;

  // Copy the five pyramid faces into the list array
  pyramid_points[Apex].pointcopy( list );
  pyramid_points[ A ].pointcopy(  list );
  pyramid_points[ B ].pointcopy(  list );
  pyramid_points[ B ].pointcopy(  list );

  pyramid_points[Apex].pointcopy( list );
  pyramid_points[ B ].pointcopy(  list );
  pyramid_points[ C ].pointcopy(  list );
  pyramid_points[ C ].pointcopy(  list );

  pyramid_points[Apex].pointcopy( list );
  pyramid_points[ C ].pointcopy(  list );
  pyramid_points[ D ].pointcopy(  list );
  pyramid_points[ D ].pointcopy(  list );

  pyramid_points[Apex].pointcopy( list );
  pyramid_points[ D ].pointcopy(  list );
  pyramid_points[ A ].pointcopy(  list );
  pyramid_points[ A ].pointcopy(  list );

  pyramid_points[ A ].pointcopy(  list );
  pyramid_points[ B ].pointcopy(  list );
  pyramid_points[ C ].pointcopy(  list );
  pyramid_points[ D ].pointcopy(  list );

} // pyr_get_faces




void display_pyramid::print_indent(int indent )
{
  int i;

  for (i = 0; i < indent; i++)
    printf(" ");
}



//
// display_pyramid::debug_print
//
// Print the pyramid data structure (e.g., the apex and the four
// corners.
//
void display_pyramid::debug_print(int level )
{

  print_indent( level );
  // print the coordinates of the apex of the pyramid
  printf("[%d, %d, %d]\n", 
	 pyramid_points[ Apex ].getX(),
	 pyramid_points[ Apex ].getY(),
	 pyramid_points[ Apex ].getZ());
  print_indent( level );
  // print the coordinates of the base of the pyramid
  printf("[%d, %d, %d] ", 
	 pyramid_points[ A ].getX(),
	 pyramid_points[ A ].getY(),
	 pyramid_points[ A ].getZ());
  printf("[%d, %d, %d] ", 
	 pyramid_points[ B ].getX(),
	 pyramid_points[ B ].getY(),
	 pyramid_points[ B ].getZ());
  printf("[%d, %d, %d] ", 
	 pyramid_points[ C ].getX(),
	 pyramid_points[ C ].getY(),
	 pyramid_points[ C ].getZ());
  printf("[%d, %d, %d]\n", 
	 pyramid_points[ D ].getX(),
	 pyramid_points[ D ].getY(),
	 pyramid_points[ D ].getZ());
}




// Static memory allocator object shared by all Sierpinski objects.
page_pool Sierpinski::mem;


//
// power_calc
//
// This function is passed a power of two.  It returns 2 ** power,
// using table look up.
//
int Sierpinski::power_calc( int power )
{
  const int power_table[MaxPower+1] = {0x001, 0x002, 0x004, 0x008, 0x010, 
                                       0x020, 0x040};
  int level;

  if (power <= MaxPower)
    level = power_table[ power ];
  else {
    printf("Sierpinski::power_calc: power is too big (power = %d)\n", power);
    exit;
  }
  return level;
} // power_calc



/*
 * calc_num_faces
 *

  The Sierpinski pyramid has four sides.  The atomic unit of the
  Sierpinski pyramid is the single four sided pyramid.  The pyramid is
  built by powers of two, doubling in size with each power.  In the
  table below, "n" is the power of two.  The height of the Sierpinski,
  in terms of atomic pyramids, is 2**n.  The number of pyramid faces
  (including the bottom of the pyramid) is 5**(n+1).  The amount of
  storage consumed by the point list array ("facelist") is also shown.
  Each face in the pyramid has four points, since the pyramid base has 
  four points and since each 3DFACE entry in the DXF file format has 
  four points.  A structure coule be used here to save some memory,
  but it would have complicated the code somewhat.  The memory
  calculation is:
  
      4 points     3 ints    4 bytes     48 bytes
      -------- X  ------- X  -------  =  --------
       face        point       int        face
   
                               storage for
         n    2**n  5**(n+1)   point list
         ---------------------------------
         0     1         5            240
         1     2        25           1200
         2     4       125           6000
         3     8       625          30000
         4    16      3125         150000
         5    32     15625         750000
         6    64     78125        3750000

  When the Sierpinski pyramid is created, its power is stored in a
  class variable.  This function returns the number of faces for
  the power of the Sierpinski pyramid.  Like the power calc function,
  this function uses table lookup rather than arithmetic calculation.

*/
int Sierpinski::calc_num_faces(void)
{
  const int MaxPower = 6;
  const int power_table[ MaxPower+1 ] = {5, 25, 125, 625, 
                                         3125, 15625, 78125};
  int num_faces = 0;

  if (pyr_power <= MaxPower )
    num_faces = power_table[ pyr_power ];
  else
    printf("display_sierpinski::calc_num_faces: 'power' is too large\n");
  return num_faces;
}  // calc_num_faces




//
// apex_calc
//
// This function is passed the three_D_point for a pyramid, the power
// for the pyramid being constructed and a number indicating a pyramid
// corner.  The function returns the three_D_point for the pyramid corner.
//
// Sierpinski pyramids are based on powers of two.  So a single pyramid
// is a power zero pyramid.  A two level Sierpinski pyramid is a power
// of one pyramid.  A four level is a power of two, and so on.
//
three_D_point Sierpinski::apex_calc( three_D_point apex, 
				     int power, 
				     int corner )
{
  int level = 1;
  int Ax, Ay, Az;
  int new_Ax = 0;
  int new_Ay = 0;
  int new_Az = 0;
  three_D_point retval;

  level = power_calc( power );
  Ax = apex.getX();
  Ay = apex.getY();
  Az = apex.getZ();
  switch (corner) {
  case 0: 
    new_Ax = Ax - level;
    new_Ay = Ay - (level * height);
    new_Az = Az - level;
    break;
  case 1:
    new_Ax = Ax + level;
    new_Ay = Ay - (level * height);
    new_Az = Az - level;
    break;
  case 2:
    new_Ax = Ax + level;
    new_Ay = Ay - (level * height);
    new_Az = Az + level;
    break;
  case 3:
    new_Ax = Ax - level;
    new_Ay = Ay - (level * height);
    new_Az = Az + level;
    break;
  }
  retval.set_point( new_Ax, new_Ay, new_Az );
  return retval;
}  // apex_calc



//
// build
//
// Recursively construct a Sierpinski pyramid.
//
// The pyramid is constructed top down.  For a power N pyramid, an
// N-1 pyramid is constructed, supported by four N-1 children.  Each
// of these pyramids is constructed in a similar way.
//
Sierpinski::sierpinski_ptr 
Sierpinski::build(three_D_point apex, int power)
{
  sierpinski_ptr new_pyramid;

  new_pyramid = (sierpinski_ptr)mem.page_alloc( sizeof( sierpinski_struct ) );

  if (power == 0) {
    int i;
    pyramid_ptr pyr_ptr;

    // allocate a pyramid object and call its constructor
    pyr_ptr = (pyramid_ptr)mem.page_alloc( sizeof( pyramid ) );
    pyr_ptr->display_pyramid::display_pyramid( apex );

    new_pyramid->obj_kind = pyramid_object;
    new_pyramid->pyr = pyr_ptr;

    // This is a building block, so the children are null
    for (i = 0; i < NumChildren; i++)
      new_pyramid->children[ i ] = NULL;
  }
  else {
    int i;

    power = power - 1;

    // build the "top" pyramid
    new_pyramid->obj_kind = sierpinski_object;
    new_pyramid->sier = build(apex, power );
    
    // build the children that support the top pyramid
    for (i = 0; i < NumChildren; i++) {
      new_pyramid->children[ i ] = build(apex_calc(apex, power, i), power );
    }
  }
  // return a pointer to the sierpinski structure
  return new_pyramid;
} // build



//
// get_faces
//
// Recursively traverse the Sierpinski pyramid and create a list
// of all the faces.  This list is stored in a "raw" integer array,
// where each integer represents a 3-D point coordinate.
//
void Sierpinski::get_faces( sierpinski_ptr root, int *&list )
{

  if (root != NULL) {
    int i;

    if (root->obj_kind == pyramid_object) {
      root->pyr->pyr_get_faces( list );
    }
    else {
      get_faces( root->sier, list);
    }
    for (i = 0; i < NumChildren; i++) {
      if (root->children[ i ] != NULL) {
	get_faces( root->children[ i ], list);
      }
    }
  }
} // Sierpinski::get_faces




//
// Sierpinski constructor
//
// This constructor is passed the "power" of the Sierpinski pyramid.
// The "power" argument controls the size of the pyramid:
//
//   power     pyramid
//   -----     -------
//     0       A single four sided pyramid
//     1       A single four sided pyramid at the apex, with
//             four single (power zero) children.
//     2       A power 1 pyramid at the apex, with four power
//             1 children.
//     N       A power N-1 pyramid at the apex, with four power
//             N-1 children.
//
Sierpinski::Sierpinski( int power )
{ 
  three_D_point apex;
  int level;

  level = power_calc( power );
  apex.set_point( level, level * height, level);

  pyr_power = power;
  root = build( apex, power );

  num_faces = calc_num_faces();
  facelist = (int *)malloc( num_faces * sizeof( face ) );

  int *tmp = facelist;

  get_faces( root, tmp);
}





//
// Top down Sierpinski print routine
//
void debug_sierpinski::sier_debug_print( sierpinski_ptr root, int level )
{
  if (root != NULL) {
    int i;

    if (root->obj_kind == pyramid_object) {
      root->pyr->debug_print(level);
    }
    else {
      level = level + 1;
      sier_debug_print( root->sier, level );
    }
    for (i = 0; i < NumChildren; i++) {
      if (root->children[i] != NULL) {
	sier_debug_print( root->children[ i ], level + 1);
      }
    }
  }
} // debug_print




void debug_sierpinski::print_facelist(void)
{
  const int size_of_point = sizeof(point)/sizeof(int);
  int i, j, k;
  int *tmp = (int *)facelist;
  
  for (k = 0; k < num_faces; k++) {
    for (j = 0; j < points_per_face; j++) {
      if (j < points_per_face-1 || point_neq( tmp, tmp - size_of_point)) {
	printf("[");
	for (i = 0; i < 3; i++) {
	  printf("%d", *tmp);
	  tmp++;
	  if (i < 2)
	    printf(",");
	} // i
	printf("]");
      }
      else
	tmp += size_of_point;
    } // j
    printf("\n");
    if ((k+1) % faces_per_pyramid == 0)
      printf("\n");
  } // k
} // print_facelist

